;;; -*- Scheme -*-

;;;; CLI interface to the PostgreSQL RDBMS for DrScheme (using libpq library)

;;; Author: Zbigniew Jurkiewicz <zbyszek@mimuw.edu.pl>
;;; Version: 0.1
;;;
;;;     Copyright (C) 2004-2006  Zbigniew Jurkiewicz

;;; This library is free software; you may redistribute it and/or
;;; modify it under the terms of the GNU Library General Public
;;; License as published by the Free Software Foundation; either
;;; version 2 of the License, or (at your option) any later version.
;;;   
;;; This library is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;; Library General Public License for more details.
;;;   
;;; You should have received a copy of the GNU Library General Public
;;; License along with this library; if not, write to the Free
;;; Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
;;;
;;; Please send suggestions and bug reports to <zbyszek@mimuw.edu.pl>
;;;
;;; The latest version of this package should be available from
;;; 
;;;     <URL:http://duch.mimuw.edu.pl/~zbyszek/>

;;; Overview =========================================================
;;;
;;; This module lets you access the PostgreSQL object-relational DBMS
;;; from DrScheme and is a simple wrapper around the libpq library.
;;;
;;; Works with MzScheme, MrEd and DrScheme.
;;;
;;; See the README for API documentation. 

(module libpq-fe mzscheme
  (require (lib "cffi.ss" "compiler"))
  (provide (all-defined))

(c-declare "#include <stdlib.h>")
(c-declare "#include <stdio.h>")
(c-declare "#include <libpq-fe.h>")
(c-declare "#include <libpq/libpq-fs.h>")

(define fopen (c-lambda (char-string char-string) (pointer "FILE") "fopen"))

(define EXIT-SUCCESS 0)
(define EXIT-FAILURE 1)

;;; ConnStatusType

(define CONNECTION-AUTH-OK
  ((c-lambda () int "___result = CONNECTION_AUTH_OK;")))
(define CONNECTION-AWAITING-RESPONSE
  ((c-lambda () int "___result = CONNECTION_AWAITING_RESPONSE;")))
(define CONNECTION-BAD ((c-lambda () int "___result = CONNECTION_BAD;")))
(define CONNECTION-MADE ((c-lambda () int "___result = CONNECTION_MADE;")))
(define CONNECTION-OK ((c-lambda () int "___result = CONNECTION_OK;")))
(define CONNECTION-SETENV ((c-lambda () int "___result = CONNECTION_SETENV;")))
(define CONNECTION-STARTED
  ((c-lambda () int "___result = CONNECTION_STARTED;")))

;;; ExecStatusType

(define PGRES-BAD-RESPONSE
  ((c-lambda () int "___result = PGRES_BAD_RESPONSE;")))
(define PGRES-COMMAND-OK ((c-lambda () int "___result = PGRES_COMMAND_OK;")))
(define PGRES-COPY-IN ((c-lambda () int "___result = PGRES_COPY_IN;")))
(define PGRES-COPY-OUT ((c-lambda () int "___result = PGRES_COPY_OUT;")))
(define PGRES-EMPTY-QUERY ((c-lambda () int "___result = PGRES_EMPTY_QUERY;")))
(define PGRES-FATAL-ERROR ((c-lambda () int "___result = PGRES_FATAL_ERROR;")))
(define PGRES-NONFATAL-ERROR
  ((c-lambda () int "___result = PGRES_NONFATAL_ERROR;")))
(define PGRES-TUPLES-OK ((c-lambda () int "___result = PGRES_TUPLES_OK;")))

;;; PostgresPollingStatusType
  
(define PGRES-POLLING-ACTIVE
  ((c-lambda () int "___result = PGRES_POLLING_ACTIVE;")))
(define PGRES-POLLING-FAILED
  ((c-lambda () int "___result = PGRES_POLLING_FAILED;")))
(define PGRES-POLLING-OK
  ((c-lambda () int "___result = PGRES_POLLING_OK;")))
(define PGRES-POLLING-READING
  ((c-lambda () int "___result = PGRES_POLLING_READING;")))
(define PGRES-POLLING-WRITING
  ((c-lambda () int "___result = PGRES_POLLING_WRITING;")))

;;;; Database Connection Functions

(define PQ-connectdb
  ;; Argument: conninfo.
  (c-lambda (char-string) (pointer "PGconn") "PQconnectdb"))

(define PQ-setdb-login
  ;; Arguments: pghost, pgport, pgoptions, pgtty, dbName, login, pwd.
  (c-lambda (char-string char-string char-string char-string char-string 
             char-string char-string) (pointer "PGconn") "PQsetdbLogin"))

(define PQ-setdb
  ;; Arguments: pghost, pgport, pgoptions, pgtty, dbName.
  (c-lambda (char-string char-string char-string char-string char-string)
            (pointer "PGconn") "PQsetdb"))

(define PQ-connect-start
  ;; Argument: conninfo.
  (c-lambda (char-string) (pointer "PGconn") "PQconnectStart"))

(define PQ-connect-poll
  ;; Argument: conn.
  ;; Result type is PostgresPollingStatusType.
  (c-lambda ((pointer "PGconn")) int "PQconnectPoll"))

(define PQ-finish
  ;; Argument: conn.
  (c-lambda ((pointer "PGconn")) void "PQfinish"))

(define PQ-reset
  ;; Argument: conn.
  (c-lambda ((pointer "PGconn")) void "PQreset"))

(define PQ-reset-start
  ;; Argument: conn.
  (c-lambda ((pointer "PGconn")) int "PQresetStart"))

(define PQ-reset-poll
  ;; Argument: conn.
  ;; Result type is PostgresPollingStatusType.
  (c-lambda ((pointer "PGconn")) int "PQresetPoll"))

(define PQ-db
  ;; Argument: conn.
  (c-lambda ((pointer "PGconn")) char-string "PQdb"))

(define PQ-user
  ;; Argument: conn.
  (c-lambda ((pointer "PGconn")) char-string "PQuser"))

(define PQ-pass
  ;; Argument: conn.
  (c-lambda ((pointer "PGconn")) char-string "PQpass"))

(define PQ-host
  ;; Argument: conn.
  (c-lambda ((pointer "PGconn")) char-string "PQhost"))

(define PQ-port
  ;; Argument: conn.
  (c-lambda ((pointer "PGconn")) char-string "PQport"))

(define PQ-tty
  ;; Argument: conn.
  (c-lambda ((pointer "PGconn")) char-string "PQtty"))

(define PQ-options
  ;; Argument: conn.
  (c-lambda ((pointer "PGconn")) char-string "PQoptions"))

(define PQ-status
  ;; Argument: conn.
  ;; Result type is ConnStatusType. 
  (c-lambda ((pointer "PGconn")) int "PQstatus")) 

(define PQ-error-message
  ;; Argument: conn
  (c-lambda ((pointer "PGconn")) char-string "PQerrorMessage"))

(define PQ-escape-string
  ;; Argument: from.
  (c-lambda (char-string) scheme-object
    "char *lto;
     Scheme_Object *lstr;

     lto = malloc(2 * strlen(___arg1) * sizeof(char) + 1);
     PQescapeString(lto, ___arg1, strlen(___arg1));
     lstr = scheme_make_string(lto);
     free(lto);
     ___result = lstr;"))

;;;; Command Execution Functions

;;; Main Routines

(define PQ-exec
  ;; Arguments: conn, query.
  (c-lambda ((pointer "PGconn") char-string) (pointer "PGresult") "PQexec"))

  (define PQ-notifies
    ;; Arguments: conn.
    (c-lambda ((pointer "PGconn")) (pointer "PGnotify") "PQnotifies"))

  (define PG-notify-relname  ;***
    (c-lambda ((pointer "PGnotify")) char-string
      "___result = ___arg1->relname;"))

  (define PQ-free-notify
    ;; Arguments: notify.
    (c-lambda ((pointer "PGnotify")) void "PQfreeNotify"))
  
  (define PQ-result-status
    ;; Argument: res
    ;; Result type is ExecStatusType.
    (c-lambda ((pointer "PGresult")) int "PQresultStatus"))

(define PQ-res-status
  ;; Argument: status (of ExecStatusType).
  (c-lambda (int) char-string "PQresStatus"))

(define PQ-result-error-message
  ;; Argument: res.
  (c-lambda ((pointer "PGresult")) char-string "PQresultErrorMessage"))

(define PQ-clear
  ;; Argument: res.
  (c-lambda ((pointer "PGresult")) void "PQclear"))

(define PQ-make-empty-pgresult
  ;; Arguments: conn, status (of ExecStatusType).
  (c-lambda ((pointer "PGconn") int) (pointer "PGresult") "PQmakeEmptyPGresult"))

;;; Retrieving SELECT Result Information

(define PQ-ntuples
  ;; Argument: res.
  (c-lambda ((pointer "PGresult")) int "PQntuples"))

(define PQ-nfields
  ;; Argument: res.
  (c-lambda ((pointer "PGresult")) int "PQnfields"))

(define PQ-fname
  ;; Arguments: res, field_index.
  (c-lambda ((pointer "PGresult") int) char-string "PQfname"))

(define PQ-fnumber
  ;; Arguments: res, field_name.
  (c-lambda ((pointer "PGresult") char-string) int "PQfnumber"))

(define PQ-ftype
  ;; Arguments: res, field_index.
  (c-lambda ((pointer "PGresult") int) int "PQftype"))

(define PQ-fmod
  ;; Arguments: res, field_index.
  (c-lambda ((pointer "PGresult") int) int "PQfmod"))

(define PQ-fsize
  ;; Arguments: res, field_index.
  (c-lambda ((pointer "PGresult") int) int "PQfsize"))

;;; Retrieving SELECT Result Values

(define PQ-getvalue
  ;; Arguments: res, tup_num, field_num.
  (c-lambda ((pointer "PGresult") int int) char-string "PQgetvalue"))

(define PQ-getisnull
  ;; Argument: res.
  (c-lambda ((pointer "PGresult") int int) bool "PQgetisnull"))

(define-struct PQ-print-opt
  (header        ;print headers (pqbool) /* print output field headings and row count */
   align         ;align fields (pqbool) /* fill align the fields */
   ;;pqbool standard;    /* old brain dead format */
   html3         ;output as html tables (pqbool)
   ;;pqbool expanded;    /* expand tables */
   ;;pqbool pager;       /* use pager for output if needed */
   field-sep     ;field separator (char*)
   ;;char *tableOpt;   /* insert to HTML <TT CLASS="REPLACEABLE"><I>table ...</I></TT> */
   ;;char *caption;    /* HTML <TT CLASS="REPLACEABLE"><I>caption</I></TT> */
   ;;char **fieldName; /* null terminated array of replacement field names */
   ;;field-name   ;alternate field names (char**)  ;???
  )
  (make-inspector))

(define (PQ-print output-stream result ps)
  ;; FILE* X PGresult* X PQprintOpt* -> void
  ((c-lambda ((pointer "FILE") (pointer "PGresult") bool bool bool char-string
              #|(pointer "char*")|#) void
     "PQprintOpt print_options;

      memset(&print_options, '\0', sizeof(print_options));
      print_options.header = ___arg3;       /* print headers */
      print_options.align = ___arg4;        /* align fields */
      print_options.html3 = ___arg5;        /* output as html tables */
      print_options.fieldSep = ___arg6;     /* field separator */
      /*print_options.fieldName = ___arg7;*/    /* alternate field names */
      PQprint(___arg1, ___arg2, &print_options);")
   output-stream result (PQ-print-opt-header ps) (PQ-print-opt-align ps)
   (PQ-print-opt-html3 ps) (PQ-print-opt-field-sep ps)
   #|(PQ-print-opt-field-name ps)|#))

  ;;; Retrieving Non-SELECT Result Information

  (define PQ-cmd-status
    ;; Argument: res.
    (c-lambda ((pointer "PGresult")) char-string "PQcmdStatus"))

  (define PQ-cmd-tuples
    ;; Argument: res.
    (c-lambda ((pointer "PGresult")) char-string "PQcmdTuples"))

  (define PQ-oid-value "not implemented yet")

  ;;; Large Objects

  (define INV-READ ((c-lambda () int "___result = INV_READ;")))
  (define INV-WRITE ((c-lambda () int "___result = INV_WRITE;")))

  (define SEEK-CUR ((c-lambda () int "___result = SEEK_CUR;")))
  (define SEEK-END ((c-lambda () int "___result = SEEK_END;")))
  (define SEEK-SET ((c-lambda () int "___result = SEEK_SET;")))

  (define lo-close
    ;; Arguments conn, fd.
    (c-lambda ((pointer "PGconn") int) int "lo_close"))

  (define lo-creat
    ;; Arguments conn, mode.
    ;; Result type is Oid. 
    (c-lambda ((pointer "PGconn") int) int "lo_creat"))

  (define lo-export
    ;; Arguments conn, lobj-id (of Oid), filename.
    (c-lambda ((pointer "PGconn") int char-string) int "lo_export"))

  (define lo-import
    ;; Arguments conn, filename.
    ;; Result type is Oid. 
    (c-lambda ((pointer "PGconn") char-string) int "lo_import"))

  (define lo-lseek
    ;; Arguments: conn, fd offset whence.
    (c-lambda ((pointer "PGconn") int int int) int "lo_lseek"))

  (define lo-open
    ;; Arguments conn, lobj-id (of Oid), mode.
    (c-lambda ((pointer "PGconn") int int) int "lo_open"))

  (define lo-read
    ;; Arguments conn, fd, buf, len (of size_t).
    (c-lambda ((pointer "PGconn") int char-string int) int "lo_read"))

  (define lo-tell
    ;; Arguments conn, fd.
    (c-lambda ((pointer "PGconn") int) int "lo_tell"))
 
  (define lo-unlink
    ;; Arguments: conn, lobj-id (of Oid).
    (c-lambda ((pointer "PGconn") int) int "lo_unlink"))

  (define lo-write
    ;; Arguments conn, fd, buf, len (of size_t).
    (c-lambda ((pointer "PGconn") int char-string int) int "lo_write")))

;;; *EOF*
