;;; Project:           Scheme-PG
;;; Author:            David J. Neu, djneu@acm.org
;;; Maintainer:        David J. Neu, djneu@acm.org
;;; Project Home Page: http://scheme-pg.sourceforge.net
;;; Copyright:         Copyright (c) 2004 Universal Technical Resource Services, Inc.
;;; License:           MIT License, see license.txt
;;; CVS Id:            $Id: stream.ss,v 1.1.1.1 2004/09/03 20:22:15 djneu Exp $

(module stream mzscheme
  (require "cursor.scm")
  (provide stream-delay
           stream-force
           stream-cons
           stream-car
           stream-cursor
           stream-cdr
           null-stream
           stream-null?
           stream->list
           stream
           stream-ref
           stream-ref!
           stream-metadata
           stream-length)

  ;; Didn't use the build-in delay and force because don't want values
  ;; to be cached.

  (define-syntax stream-delay
    (syntax-rules ()
      ((_ expr) (lambda () expr))))

  (define (stream-force apromise)
    (apromise))

  ;; Returns a pair with the car being a value and the cdr being a
  ;; promise that when forced returns the next pair in the stream.

  (define-syntax stream-cons
    (syntax-rules ()
      ((_ avalue astream) (cons avalue (stream-delay astream)))))

  (define (stream-car astream)
    ;;(car astream))
    (caar astream))

  (define (stream-cursor astream)
    ;(car astream))
    (cdar astream))

  (define (stream-cdr astream)
    (stream-force (cdr astream)))

  (define null-stream '())

  (define (stream-null? astream)
    (null? astream))

  ;; Returns the elements of the stream as a list.

  (define (stream->list astream)
    (if (stream-null? astream)
        '()
        (cons (stream-car astream) (stream->list (stream-cdr astream)))))

  ;; Creates a cursor and returns a stream that contains it's contents.

  (define (stream aconnection aselect aclient/server)
    (let loop ((lcursor (cursor-open aconnection aselect aclient/server)))
      (let ((lrow (cursor-next lcursor)))
        (if (eoc-object? lrow)
            (begin
              (cursor-close lcursor)
              null-stream)
            ;;(stream-cons lrow (loop lcursor)))))))
            (stream-cons (cons lrow lcursor) (loop lcursor))))))

  (define (stream-ref astream aindex)
    (let ((lcursor (stream-cursor astream)))
      (cursor-ref lcursor aindex)))

  (define (stream-ref! astream aindex)
    (let ((lcursor (stream-cursor astream)))
      (cursor-ref! lcursor aindex)))

  (define (stream-metadata astream aindex)
    (let ((lcursor (stream-cursor astream)))
      (cursor-metadata lcursor)))

  ;; Since all the streams we work with are finite, we can provide the
  ;; stream-length function.

  (define (stream-length astream aindex)
    (let ((lcursor (stream-cursor astream)))
      (cursor-length lcursor))))

;;; *EOF*
