;;;; Scheme Language Interface (libpq) sample program

(require (lib "libpq-fe.scm" "libpq"))
(require (lib "etc.ss"))

(define conn #f)

(define (test-cursor)
  (let ((connection-str "dbname=bd user=scott password=tiger"))  ;host=labdb 
    (set! conn (PQ-connectdb connection-str))
    (cond ((= (PQ-status conn) CONNECTION-BAD)
           (fprintf (current-error-port)
                    "Connection to ~a failed, ~a" connection-str
                    (PQ-error-message conn)))
          (else
           (printf "Connected OK~%")
           (let ((result (execute-one-statement "BEGIN WORK")))
             (when result
               (PQ-clear result)
               (set! result
                     (execute-one-statement
                      "DECLARE przysmak_cursor CURSOR FOR SELECT gatunek, przysmak FROM Gatunki WHERE kontynent = 'Europa'"))
               (when result
                 (PQ-clear result)
                 (set! result
                       (execute-one-statement "FETCH 1 IN przysmak_cursor"))
                 (when result (show-column-info result))
                 (let loop ()
                   (when (and result (> (PQ-ntuples result) 0))
                     (show-one-row-data result)
                     (PQ-clear result)
                     (set! result
                           (execute-one-statement "FETCH NEXT IN przysmak_cursor"))
                     (loop)))
                 (set! result
                       (execute-one-statement "COMMIT WORK"))
                 (when result (PQ-clear result)))))))
    (when conn (PQ-finish conn))))

(define (execute-one-statement stmt-to-exec)
  (printf "About to execute ~a~%" stmt-to-exec)
  (let ((result (PQ-exec conn stmt-to-exec)))
    (cond ((not result)
           (printf "PQ-exec command failed, no error code~%")
           #f)
          (else
           (evcase (PQ-result-status result)
             (PGRES-COMMAND-OK 
              (let ((str-res (PQ-cmd-tuples result)))
                (if (> (string-length str-res) 0)
                    (printf "Command executed OK, ~a rows affected~%"
                            str-res)
                    (printf "Command executed OK, no rows affected~%"))
                result))
             (PGRES-TUPLES-OK
              (printf "Select executed OK, ~a rows found~%"
                      (PQ-ntuples result))
              result)
             (else
              (printf "Command failed with code ~a, error message ~a~%"
                      (PQ-res-status (PQ-result-status result))
                      (PQ-result-error-message result))
              (PQ-clear result)
              #f))))))

(define (show-column-info result)
  (when result
    (let ((num-columns (PQ-nfields result)))
      (printf "~a Columns in the result~%" num-columns)
      (do ((i 0 (+ i 1)))
        ((>= i num-columns))
        (printf "Field ~a, Name ~a, Internal size ~a~%" 
                i (PQ-fname result i) (PQ-fsize result i))))))

(define (show-one-row-data result)
  (do ((col 0 (+ col 1)))
    ((>= col (PQ-nfields result) 2))
    (printf "DATA: ~a~%"
            (if (PQ-getisnull result 0 col)
                "<NULL>"
                (PQ-getvalue result 0 col)))))

;;; *EOF*













